"""MOTCO package.

Multi-omics Trajectory Comparison utilities:
- Latent space generation (PLS-DA, SNF)
- Group differences on multivariate trajectories
"""

__all__ = [
    "stats",
]

__version__ = "0.1.1"
